const { expect } = require('chai')
const sinon = require('sinon')
const Injectable = require('../src/Injectable')

describe('the injectable base class', function() {
    it('should be a functional class', function() {
        expect(Injectable).to.be.a('function')
    })

    it('should be constructable', function() {
        expect(new Injectable()).to.be.an.instanceof(Injectable)
    })

    it('should have a _di_allow_defer field', function() {
        expect(Injectable._di_allow_defer).to.not.be.undefined
    })

    it('should have a _di_deferred_services array', function() {
        expect(Injectable._di_deferred_services).to.be.an('array')
    })

    it('should have a _di_deferred_instances array', function() {
        expect(Injectable._di_deferred_instances).to.be.an('array')
    })

    it('should have a services array', function() {
        expect(Injectable.services).to.be.an('array')
    })

    it('should have __has_deferred_services property', function() {
        expect(Injectable.__has_deferred_services).to.not.be.undefined
    })

    describe('__has_deferred_services property', function() {
        it('should exist', function() {
            expect(Injectable.__has_deferred_services).to.not.be.undefined
        })

        it('should return false if there are no _di_deferred_services', function() {
            Injectable._di_deferred_services = []
            expect(Injectable.__has_deferred_services).to.be.false
        })

        it('should return true if there are _di_deferred_services', function() {
            Injectable._di_deferred_services = ['some service']
            expect(Injectable.__has_deferred_services).to.be.true
            Injectable._di_deferred_services = []
        })
    })

    describe('__inject function', function() {
        let InjClass
        let svc_name
        let FakeService
        let fake_container
        beforeEach(function() {
            svc_name = Math.random().toString(36).substring(7)
            class Test extends Injectable {
                static get services() {
                    return [svc_name]
                }
            }
            InjClass = Test
            class SomeService {}
            FakeService = SomeService
            fake_container = {
                get: sinon.fake.returns(new FakeService)
            }
        })

        it('should request the services from the container, if deferral is disabled', function() {
            InjClass._di_allow_defer = false
            InjClass.__inject(fake_container)
            expect(fake_container.get.calledOnce).to.be.true
            expect(fake_container.get.getCall(0).args[0]).to.be.equal(svc_name)
        })

        it('should not defer the services if they do exist', function() {
            InjClass._di_allow_defer = true
            fake_container.has = sinon.fake.returns(true)
            fake_container.defer = sinon.spy()
            fake_container.get = sinon.spy()
            InjClass.__inject(fake_container)
            expect(fake_container.has.calledOnce).to.be.true
            expect(InjClass._di_deferred_services).to.not.include(svc_name)
            expect(InjClass._di_deferred_services.length).to.be.equal(0)
            expect(fake_container.defer.calledOnce).to.be.false
            expect(fake_container.get.calledOnce).to.be.true
            expect(fake_container.get.getCall(0).args[0]).to.be.equal(svc_name)
        })

        it('should add the services to the class prototype from the container', function() {
            InjClass._di_allow_defer = false
            InjClass.__inject(fake_container)
            expect(InjClass.prototype[svc_name]).to.be.an.instanceof(FakeService)
        })

        it('should defer the services if they do not exist', function() {
            InjClass._di_allow_defer = true
            fake_container.has = sinon.fake.returns(false)
            fake_container.defer = sinon.spy()
            InjClass.__inject(fake_container)
            expect(fake_container.has.calledOnce).to.be.true
            expect(InjClass._di_deferred_services).to.include(svc_name)
            expect(fake_container.defer.calledOnce).to.be.true
            expect(fake_container.defer.getCall(0).args[0]).to.be.equal(InjClass)
        })
    })

    describe('__deferral_callback function', function() {
        it('should be a function', function() {
            expect(Injectable.__deferral_callback).to.be.a('function')
        })

        it('should expect 2 arguments', function() {
            expect(Injectable.__deferral_callback.length).to.be.equal(2)
        })

        it('should inject the instance into deferred classes', function() {
            const name = Math.random().toString(36).substring(7)
            const fake_class = {}
            class InjClass extends Injectable {
                static _di_deferred_services = [name]
                static _di_deferred_instances = [fake_class]
            }

            const fake_instance = {}
            InjClass.__deferral_callback(name, fake_instance)
            expect(fake_class[name]).to.be.equal(fake_instance)
        })

        it('should remove the service from deferred services once fulfilled', function() {
            const name = Math.random().toString(36).substring(7)
            const fake_class = {}
            class InjClass extends Injectable {
                static _di_deferred_services = [name]
                static _di_deferred_instances = [fake_class]
            }

            const fake_instance = {}
            InjClass.__deferral_callback(name, fake_instance)
            expect(InjClass._di_deferred_services).to.not.include(name)
        })

        it('should do nothing if there are no deferred services', function() {
            const name = Math.random().toString(36).substring(7)
            class InjClass extends Injectable {
                static _di_deferred_services = []
                static _di_deferred_instances = []
            }

            const fake_instance = {}
            InjClass.__deferral_callback(name, fake_instance)
        })
    })

    describe('constructor', function() {
        it('should add instances to deferred instances if there are deferred services', function() {
            class TestClass extends Injectable {
                static _di_allow_defer = true
                static get __has_deferred_services() {
                    return true
                }
            }

            const tc = new TestClass()
            expect(TestClass._di_deferred_instances).to.include(tc)
        })
    })
})
